"""The module allows to run tests from inside the interpreter."""
import os

import pytest

from grmpy.simulate.simulate import simulate
from grmpy.estimate.estimate import estimate


def test():
    """The function allows to run the tests from inside the interpreter."""

    package_directory = os.path.dirname(os.path.realpath(__file__))
    current_directory = os.getcwd()

    os.chdir(package_directory)

    pytest.main()

    os.chdir(current_directory)



""" This module provides auxiliary functions for the simulate.py module. It includes
simulation processes of the unobservable and endogenous variables of the model as well
as functions regarding the info file output.
"""
import numpy as np
import pandas as pd
from scipy.stats import norm


def simulate_covariates(init_dict):
    """The function simulates the covariates for the choice and the output functions."""
    # Distribute information
    num_agents = init_dict["SIMULATION"]["agents"]

    # Construct auxiliary information

    num_covars = init_dict["AUX"]["num_covars"]
    labels = init_dict["AUX"]["labels"]

    # As our baseline we simulate covariates from a standard normal distribution.
    means = np.tile(0.0, num_covars)
    covs = np.identity(num_covars)
    X = pd.DataFrame(
        np.random.multivariate_normal(means, covs, num_agents), columns=labels
    )

    # We now perform some selective replacements.
    # Set intercept
    X[labels[0]] = 1.0

    # Include binary variables
    if "VARTYPES" in init_dict:
        for variable in init_dict["VARTYPES"]:
            if isinstance(init_dict["VARTYPES"][variable], list):
                X[variable] = np.random.binomial(
                    1, init_dict["VARTYPES"][variable][1], size=num_agents
                )
    else:
        pass

    return X


def simulate_unobservables(init_dict):
    """The function simulates the unobservable error terms."""
    num_agents = init_dict["SIMULATION"]["agents"]
    cov = construct_covariance_matrix(init_dict)

    U = pd.DataFrame(
        np.random.multivariate_normal(np.zeros(3), cov, num_agents),
        columns=["U1", "U0", "V"],
    )

    return U


def simulate_outcomes(init_dict, X, U):
    """The function simulates the potential outcomes Y0 and Y1, the resulting treatment
    dummy D and the realized outcome Y.
    """
    dep = init_dict["ESTIMATION"]["dependent"]
    indicator = init_dict["ESTIMATION"]["indicator"]

    df = X.join(U)
    Z = df[init_dict["CHOICE"]["order"]].values
    X_treated, X_untreated = (
        df[init_dict["TREATED"]["order"]],
        df[init_dict["UNTREATED"]["order"]],
    )

    # Distribute information
    coeffs_untreated = init_dict["UNTREATED"]["params"]
    coeffs_treated = init_dict["TREATED"]["params"]
    coeffs_choice = init_dict["CHOICE"]["params"]

    # Calculate potential outcomes and choice
    df[dep + "1"] = np.dot(coeffs_treated, X_treated.T) + df["U1"]
    df[dep + "0"] = np.dot(coeffs_untreated, X_untreated.T) + df["U0"]
    C = np.dot(coeffs_choice, Z.T) - df["V"]

    # Calculate expected benefit and the resulting treatment dummy
    df[indicator] = np.array((C > 0).astype(float))

    # Observed outcomes
    df[dep] = df[indicator] * df[dep + "1"] + (1 - df[indicator]) * df[dep + "0"]

    return df


def write_output(init_dict, df):
    """The function converts the simulated variables to a panda data frame and saves the
    data in a txt and a pickle file.
    """
    # Distribute information
    source = init_dict["SIMULATION"]["source"]

    df.to_pickle(source + ".grmpy.pkl")
    with open(source + ".grmpy.txt", "w") as file_:
        df.to_string(file_, index=False, na_rep=".", col_space=15, justify="left")
    return df


def print_info(init_dict, data_frame):
    """The function writes an info file for the specific data frame."""
    # Distribute information
    coeffs_untreated = init_dict["UNTREATED"]["params"]
    coeffs_treated = init_dict["TREATED"]["params"]
    source = init_dict["SIMULATION"]["source"]
    dep, indicator = (
        init_dict["ESTIMATION"]["dependent"],
        init_dict["ESTIMATION"]["indicator"],
    )

    # Construct auxiliary information
    cov = construct_covariance_matrix(init_dict)

    with open(source + ".grmpy.info", "w") as file_:

        # First we note some basic information ab out the dataset.
        header = "\n\n Number of Observations \n\n"
        file_.write(header)

        info_ = [
            data_frame.shape[0],
            (data_frame[indicator] == 1).sum(),
            (data_frame[indicator] == 0).sum(),
        ]

        fmt = "  {:<10}" + " {:>20}" * 1 + "\n\n"
        file_.write(fmt.format(*["", "Count"]))

        for i, label in enumerate(["All", "Treated", "Untreated"]):
            str_ = "  {:<10} {:20}\n"
            file_.write(str_.format(*[label, info_[i]]))

        # Second, we describe the distribution of outcomes and effects.
        for label in ["Outcomes", "Effects"]:

            header = "\n\n Distribution of " + label + "\n\n"
            file_.write(header)

            fmt = "  {:<10}" + " {:>20}" * 5 + "\n\n"
            args = ["", "Mean", "Std-Dev.", "25%", "50%", "75%"]
            file_.write(fmt.format(*args))

            for group in ["All", "Treated", "Untreated"]:

                if label == "Outcomes":
                    data = data_frame[dep]
                elif label == "Effects":
                    data = data_frame[dep + "1"] - data_frame[dep + "0"]

                if group == "Treated":
                    data = data[data_frame[indicator] == 1]
                elif group == "Untreated":
                    data = data[data_frame[indicator] == 0]
                else:
                    pass
                fmt = "  {:<10}" + " {:>20.4f}" * 5 + "\n"
                info = list(data.describe().tolist()[i] for i in [1, 2, 4, 5, 6])
                if pd.isnull(info).all():
                    fmt = "  {:<10}" + " {:>20}" * 5 + "\n"
                    info = ["---"] * 5
                elif pd.isnull(info[1]):
                    info[1] = "---"
                    fmt = "  {:<10}" " {:>20.4f}" " {:>20}" + " {:>20.4f}" * 3 + "\n"

                file_.write(fmt.format(*[group] + info))

        # Implement the criteria function value , the MTE and parameterization
        header = "\n\n {} \n\n".format("Criterion Function")
        file_.write(header)
        if "criteria_value" in init_dict["AUX"].keys():
            str_ = "  {:<10}      {:<21.12f}\n\n".format(
                "Value", init_dict["AUX"]["criteria_value"]
            )
        else:
            str_ = "  {:>10} {:>20}\n\n".format("Value", "---")
        file_.write(str_)

        header = "\n\n {} \n\n".format("Marginal Treatment Effect")
        file_.write(header)
        quantiles = [1] + np.arange(5, 100, 5).tolist() + [99]
        args = [str(i) + "%" for i in quantiles]
        quantiles = [i * 0.01 for i in quantiles]

        x = data_frame
        value = mte_information(
            coeffs_treated, coeffs_untreated, cov, quantiles, x, init_dict
        )
        str_ = "  {:>10} {:>20}\n\n".format("Quantile", "Value")
        file_.write(str_)
        len_ = len(value)
        for i in range(len_):
            if isinstance(value[i], float):
                file_.write(f"  {str(args[i]):>10} {value[i]:>20.4f}\n")
            else:
                file_.write(f"  {str(args[i]):>10} {value[i]:>20.4}\n")

        # Write out parameterization of the model.
        write_identifier_section_simulate(init_dict, file_)


def mte_information(coeffs_treated, coeffs_untreated, cov, quantiles, x, dict_):
    """The function calculates the marginal treatment effect for pre specified quantiles
    of the collected unobservable variables.
    """

    labels = [k for k in dict_["TREATED"]["order"]]
    labels += [j for j in dict_["UNTREATED"]["order"] if j not in labels]

    # Construct auxiliary information
    if dict_["TREATED"]["order"] != dict_["UNTREATED"]["order"]:

        para_diff = []
        for var in labels:
            if var in dict_["TREATED"]["order"] and var in dict_["UNTREATED"]["order"]:
                index_treated = dict_["TREATED"]["order"].index(var)
                index_untreated = dict_["UNTREATED"]["order"].index(var)
                diff = (
                    dict_["TREATED"]["params"][index_treated]
                    - dict_["UNTREATED"]["params"][index_untreated]
                )
            elif (
                var in dict_["TREATED"]["order"]
                and var not in dict_["UNTREATED"]["order"]
            ):
                index = dict_["TREATED"]["order"].index(var)
                diff = dict_["TREATED"]["params"][index]

            elif (
                var not in dict_["TREATED"]["order"]
                and var in dict_["UNTREATED"]["order"]
            ):
                index = dict_["UNTREATED"]["order"].index(var)
                diff = -dict_["UNTREATED"]["params"][index]
            para_diff += [diff]
    else:

        para_diff = coeffs_treated - coeffs_untreated
    x = x[labels]
    MTE = []
    for i in quantiles:
        if cov[2, 2] == 0.00:
            MTE += ["---"]
        else:
            MTE += [
                np.mean(np.dot(x, para_diff)) + (cov[2, 0] - cov[2, 1]) * norm.ppf(i)
            ]

    return MTE


def write_identifier_section_simulate(init_dict, file_):
    """This function prints the information about the estimation results in the output
    file.
    """

    file_.write("\n\n {} \n\n".format("Parameterization"))

    fmt_ = " {:<10}" + "    {:>10}" + "{:>15}"

    file_.write(fmt_.format(*["Section", "Identifier", "Coef"]) + "\n")

    num_treated = len(init_dict["TREATED"]["order"])
    num_untreated = num_treated + len(init_dict["UNTREATED"]["order"])
    num_choice = num_untreated + len(init_dict["CHOICE"]["order"])

    identifier_treated = init_dict["TREATED"]["order"]
    identifier_untreated = init_dict["UNTREATED"]["order"]
    identifier_choice = init_dict["CHOICE"]["order"]
    identifier_distribution = [
        "sigma1",
        "sigma10",
        "sigma1v",
        "sigma0",
        "sigma0v",
        "sigmaV",
    ]
    identifier = (
        identifier_treated
        + identifier_untreated
        + identifier_choice
        + identifier_distribution
    )
    coeffs = init_dict["AUX"]["init_values"].copy()
    fmt = "  {:>10}" + "   {:<15}" + " {:>11.4f}"
    for i in range(len(init_dict["AUX"]["init_values"])):
        if i == 0:
            file_.write("\n  {:<10} \n".format("TREATED"))
        elif i == num_treated:
            file_.write("\n  {:<10} \n".format("UNTREATED"))
        elif i == num_untreated:
            file_.write("\n  {:<10} \n".format("CHOICE"))
        elif i == num_choice:
            file_.write("\n  {:<10} \n".format("DIST"))

        file_.write("{}\n".format(fmt.format("", identifier[i], coeffs[i])))


def construct_covariance_matrix(init_dict):
    """This function constructs the covariance matrix based on the user's initialization
    file.
    """
    cov = np.zeros((3, 3))
    cov[np.triu_indices(3)] = init_dict["DIST"]["params"]
    cov[np.tril_indices(3, k=-1)] = cov[np.triu_indices(3, k=1)]
    cov[np.diag_indices(3)] **= 2
    return cov



"""This module contains methods for producing the estimation output files."""
import time
from textwrap import wrap


def print_logfile(init_dict, rslt, print_output):
    """The function writes the log file for the estimation process."""
    # Adjust output
    if "output_file" in init_dict["ESTIMATION"].keys():
        file_name = init_dict["ESTIMATION"]["output_file"]
    else:
        file_name = "est.grmpy.info"
    file_input = ""

    # obtain optimization information and results
    opt_info = rslt["opt_info"]
    opt_rslt = rslt["opt_rslt"]

    # create header of output table
    header = "{:>50}".format("Optimization Results") + "\n" + 80 * "=" + "\n"

    # specify some table inputs
    message = wrap(opt_info["message"], 24)
    if len(message) == 1:
        message += [""]
    time_now = time.localtime()
    time_ = time.strftime("%H:%M:%S", time_now)
    date = time.strftime("%a, %d %b %Y", time_now)

    # define correctly aligned string
    fmt = "{:<15}" + "{:>20}" + "{:>5}" + "{:<16}" + "{:>24}\n"

    # define table input
    header_input = [
        ("Dep. Variable:", opt_info["dependent"], "Optimizer:", opt_info["optimizer"]),
        ("Choice Var:", opt_info["indicator"], "No. Evaluations:", opt_info["nit"]),
        ("Date:", date, "Success:", opt_info["success"]),
        ("Time:", time_, "Status:", opt_info["status"]),
        ("Observations:", opt_info["observations"], "Message:", message[0]),
        ("Start Values:", opt_info["start"], "", message[1]),
    ]

    # loop over header input

    for line in header_input:
        header += fmt.format(line[0], line[1], "", line[2], line[3])

    header += fmt.format("", "", "", "Criterion Func:", "")
    fmt = "{:<15}" + "{:>20}" + "{:>10}" + "{:<16}" + "{:>+19.4f}\n"
    for section in [
        ("Start", init_dict["AUX"]["criteria"]),
        ("Finish", opt_info["crit"]),
    ]:
        header += fmt.format("", "", "", section[0] + ":", section[1])

    header += "=" * 80 + "\n"

    # Add estimation output
    fmt = "{:<17}" + "{:>11}" + "{:>10}" * 3 + "{:>11}" + "{:>11}" "\n"

    header += fmt.format("", "coef", "std err", "t", "P>|t|", "[0.025", "0.975]")
    header += "-" * 80

    estimation_output = write_identifier_section(opt_rslt, opt_info)

    file_input = header + estimation_output

    if print_output is True:
        print(file_input)
    with open(file_name, "w") as file_:
        file_.write(file_input)


def write_identifier_section(opt_rslt, opt_info):
    """This function prints the information about the estimation results in the output
    file.
    """

    # write estimation output
    est_out = ""
    fmt_section = "\n{:<14}\n\n"
    fmt = (
        "{:<17.17}"
        + "{:>11.4f}"
        + "{:>10.3f}"
        + "{:>10.3f}"
        + "{:>10.3f}"
        + "{:>11.3f}" * 2
        + "\n"
    )
    for section in ["TREATED", "UNTREATED", "CHOICE", "DIST"]:
        est_out += fmt_section.format(section)
        params = opt_rslt.loc[section, "params"]
        std = opt_rslt.loc[section, "std"]
        t = opt_rslt.loc[section, "t_values"]
        P = opt_rslt.loc[section, "p_values"]
        conf_int_low = opt_rslt.loc[section, "conf_int_low"]
        conf_int_up = opt_rslt.loc[section, "conf_int_up"]

        for var_label in opt_rslt.loc[section, slice(None)].index:
            est_out += fmt.format(
                var_label,
                params.loc[var_label],
                std.loc[var_label],
                t.loc[var_label],
                P.loc[var_label],
                conf_int_low.loc[var_label],
                conf_int_up.loc[var_label],
            )
    est_out += 80 * "=" + "\n"

    # Add warnings to table

    est_out += "\n{}\n".format("Warning:")

    for i in opt_info["warning"]:
        est_out += "\n"
        lines = wrap(i, 80)
        for j in lines:
            est_out += f"{j}\n"

    return est_out



"""
This module provides auxiliary functions for the plot_mte function.
"""
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import norm
from sklearn.utils import resample

from grmpy.check.auxiliary import read_data
from grmpy.check.check import check_presence_init
from grmpy.estimate.estimate_semipar import (
    double_residual_reg,
    estimate_treatment_propensity,
    mte_observed,
    mte_unobserved_semipar,
    process_primary_inputs,
    process_secondary_inputs,
    trim_support,
)
from grmpy.read.read import read

# surpress pandas warning
pd.options.mode.chained_assignment = None


def plot_curve(mte, quantiles, con_u, con_d, font_size, label_size, color, save_plot):
    """
    This function plots the MTE curve (based on either the
    parametric or semiparmaetric model) along with its
    90 percent confidence bands.

    Parameters
    ----------
    mte: np.ndarray
        Estimate of the parametric MTE.
    quantiles: np.ndarray
        Quantiles of the u_D, along which the *mte* has been estimated.
    con_u: list
        Upper bound of the 90 percent confidence interval.
    con_d: list
        Lower bound of the 90 percent confidence interval.
    font_size: int, default is 22
        Font size of the MTE graph.
    label_size: int, default is 16
        Label size of the MTE graph
    color: str, default is "blue"
        Color of the MTE curve.
    save_plot: bool or str or PathLike or file-like object, default is False
        If False, the resulting plot is shown but not saved.
        If True, the MTE plot is saved as 'MTE_plot.png'.
        Else, if a str or Pathlike or file-like object is specified,
        the plot is saved according to *save_plot*.
        The output format is inferred from the extension ('png', 'pdf', 'svg'... etc.)
        By default, '.png' is assumed.
    """
    ax = plt.figure(figsize=(17.5, 10)).add_subplot(111)

    ax.set_ylabel(r"$MTE$", fontsize=font_size)
    ax.set_xlabel("$u_D$", fontsize=font_size)
    ax.tick_params(
        axis="both",
        direction="in",
        length=5,
        width=1,
        grid_alpha=0.25,
        labelsize=label_size,
    )
    # TODO: trim the y axis based on the results
    ax.xaxis.set_ticks_position("both")
    ax.yaxis.set_ticks_position("both")

    ax.plot(quantiles, mte, color=color, linewidth=4)
    ax.plot(quantiles, con_u, color=color, linestyle=":", linewidth=3)
    ax.plot(quantiles, con_d, color=color, linestyle=":", linewidth=3)

    if save_plot is False:
        pass
    elif save_plot is True:
        plt.savefig("MTE_plot.png", dpi=300)
    else:
        plt.savefig(save_plot, dpi=300)

    plt.show()


def mte_and_cof_int_semipar(rslt, init_file, college_years, nboot):
    """
    This function returns the semiparametric MTE divided by the number
    of college years, which represents the returns per YEAR of
    post-secondary schooling.
    The corresponding 90 percent confidence intervals are bootstrapped
    based on *nboot* iterations.

    Parameters
    ----------
    rslt: dict
        Result dictionary containing parameters for the estimation
        process.
    init_file: yaml
        Initialization file containing parameters for the estimation
        process.
    college_years: int, default is 4
        Average duration of college degree. The MTE plotted will thus
        refer to the returns per one year of college education.
    nboot: int
        Number of bootstrap iterations, i.e. number of times
        the MTE is computed via bootstrap.

    Returns
    -------
    quantiles: np.ndarray
        Quantiles of the u_D, along which the *mte* has been estimated.
    mte: np.ndarray
        Estimate of the parametric MTE.
    con_u: list
        Upper bound of the 90 percent confidence interval.
    con_d: list
        Lower bound of the 90 percent confidence interval.
    """
    # Define quantiles of u_D (unobserved resistance to treatment)
    quantiles = rslt["quantiles"]

    # MTE per year of post-secondary education
    mte = rslt["mte"] / college_years

    # bootstrap 90 percent confidence bands
    mte_boot = bootstrap(init_file, nboot)
    mte_boot = mte_boot / college_years

    # Get standard error of MTE at each gridpoint u_D
    mte_boot_std = np.std(mte_boot, axis=1)

    # Compute 90 percent confidence intervals
    con_u = mte + norm.ppf(0.95) * mte_boot_std
    con_d = mte - norm.ppf(0.95) * mte_boot_std

    return quantiles, mte, con_u, con_d


def mte_and_cof_int_par(rslt, data, college_years):
    """
    This function returns the parametric MTE divided by the number
    of college years, which represents the returns per YEAR of
    post-secondary schooling.
    90 percent confidence intervals are computed analytically.

    Parameters
    ----------
    rslt: dict
        Result dictionary containing parameters for the estimation
        process.
    data: pandas.DataFrame
        Data set containing the observables (explanatory and outcome variables)
        analyzed in the generalized Roy framework.
    college_years: int, default is 4
        Average duration of college degree. The MTE plotted will thus
        refer to the returns per one year of college education.
    """
    # Define quantiles of u_D (unobserved resistance to treatment)
    quantiles = rslt["quantiles"]

    # MTE per year of post-secondary education
    mte = rslt["mte"] / college_years

    con_u, con_d = calculate_cof_int(rslt, data, quantiles)

    return quantiles, mte, mte + con_u / college_years, mte - con_d / college_years


def calculate_cof_int(rslt, data, quantiles):
    """
    This function calculates the analytical confidence intervals of
    the parametric marginal treatment effect.

    Parameters
    ----------
    rslt: dict
        Result dictionary containing parameters for the estimation
        process.
    data: pandas.DataFrame
        Data set containing the observables (explanatory and outcome variables)
        analyzed in the generalized Roy framework.
    quantiles: np.ndarray
        Quantiles of the u_D, along which the *mte* has been estimated.

    Returns
    ------
    con_u: list
        Upper bound of the 90 percent confidence interval.
    con_d: list
        Lower bound of the 90 percent confidence interval.
    """

    # Import parameters and inverse hessian matrix
    hess_inv = rslt["hessian_inv"] / data.shape[0]
    params = rslt["opt_rslt"]["params"].values
    numx = (
        rslt["opt_rslt"].loc["TREATED"].shape[0]
        + rslt["opt_rslt"].loc["UNTREATED"].shape[0]
    )

    # Distribute parameters
    dist_cov = hess_inv[-4:, -4:]
    param_cov = hess_inv[:numx, :numx]
    dist_gradients = params[-4:]

    # Process data

    # goal should be to take into account that the treated and the
    # untreated section can contain different covariates
    x_treated = np.mean(data[rslt["opt_rslt"].loc["TREATED"].index.values]).values
    x_untreated = np.mean(data[rslt["opt_rslt"].loc["UNTREATED"].index.values]).values
    x = np.append(x_treated, -x_untreated)

    # Create auxiliary parameters
    part1 = np.dot(x, np.dot(param_cov, x))
    part2 = np.dot(dist_gradients, np.dot(dist_cov, dist_gradients))

    # Prepare two lists for storing the values

    # Combine all auxiliary parameters and calculate the confidence intervals
    value = part2 * (norm.ppf(quantiles)) ** 2
    aux = np.sqrt(part1 + value)
    con_u = norm.ppf(0.95) * aux
    con_d = norm.ppf(0.95) * aux

    return con_u, con_d


def bootstrap(init_file, nboot):
    """
    This function generates bootsrapped standard errors
    given an init_file and the number of bootstraps to be drawn.

    Parameters
    ----------
    init_file: yaml
        Initialization file containing parameters for the estimation
        process.
    nboot: int
        Number of bootstrap iterations, i.e. number of times
        the MTE is computed via bootstrap.

    Returns
    -------
    mte_boot: np.ndarray
        Array containing *nbootstrap* estimates of the MTE.
    """
    check_presence_init(init_file)
    dict_ = read(init_file, semipar=True)

    # Process the information specified in the initialization file
    bins, logit, bandwidth, gridsize, startgrid, endgrid = process_primary_inputs(dict_)
    trim, rbandwidth, reestimate_p, show_output = process_secondary_inputs(dict_)

    # Suppress output
    show_output = False

    # Prepare empty array to store output values
    mte_boot = np.zeros([gridsize, nboot])

    # Load the baseline data
    data = read_data(dict_["ESTIMATION"]["file"])

    counter = 0
    while counter < nboot:
        boot_data = resample(data, replace=True, n_samples=len(data), random_state=None)

        # Estimate propensity score P(z)
        boot_data = estimate_treatment_propensity(dict_, boot_data, logit, show_output)
        prop_score = boot_data["prop_score"]
        if isinstance(prop_score, pd.Series):
            # Define common support and trim the data (if trim=True)
            X, Y, prop_score = trim_support(
                dict_, boot_data, logit, bins, trim, reestimate_p, show_output=False
            )

            b0, b1_b0 = double_residual_reg(X, Y, prop_score)

            # # Construct the MTE
            mte_x = mte_observed(X, b1_b0)
            mte_u = mte_unobserved_semipar(
                X, Y, b0, b1_b0, prop_score, bandwidth, gridsize, startgrid, endgrid
            )

            # Put the MTE together
            mte = mte_x.mean(axis=0) + mte_u
            mte_boot[:, counter] = mte

            counter += 1

        else:
            continue

    return mte_boot



"""This module provides some configuration for the package."""
import os
import sys
import warnings

import numpy as np
from pathlib import Path
from statsmodels.tools.sm_exceptions import ConvergenceWarning

# Obtain the root directory of the package.
# Do not import grmpy which creates a circular import.
ROOT_DIR = Path(__file__).parent


# Directory with additional resources for the testing harness
TEST_DIR = ROOT_DIR / "tests"
TEST_RESOURCES_DIR = ROOT_DIR / "tests" / "resources"

# Set Debug mode dependend on the environment in which the test battery is running
IS_PRODUCTION = False

if os.getenv("GRMPY_DEV") == "TRUE":
    IS_PRODUCTION = True

# We want to turn off selected warnings.
if IS_PRODUCTION is False:
    warnings.simplefilter(action="ignore", category=FutureWarning)
    warnings.simplefilter(action="ignore", category=ConvergenceWarning)
    warnings.simplefilter(action="ignore", category=RuntimeWarning)

# We only support modern Python.
np.testing.assert_equal(sys.version_info[0], 3)
np.testing.assert_equal(sys.version_info[1] >= 5, True)

# We rely on relative paths throughout the package.
PACKAGE_DIR = os.path.dirname(os.path.abspath(__file__))
TEST_RESOURCES_DIR = PACKAGE_DIR + "/test/resources"



"""This is the entry-point to the grmpy package.

Include only imports which should be available using

.. code-block::

    import grmpy as gp

    gp.<func>
"""
import pytest

from grmpy.estimate.estimate import fit  # noqa: F401
from grmpy.grmpy_config import ROOT_DIR
from grmpy.plot.plot import plot_mte  # noqa: F401
from grmpy.simulate.simulate import simulate  # noqa: F401


def test(*args, **kwargs):
    """Run basic tests of the package."""
    pytest.main([str(ROOT_DIR), *args], **kwargs)



"""This module contains a monte carlo example that illustrates the advantages of the grmpy estima-
tion strategy. For this purpose data and the associated parameterization from Cainero 2011 are
used. Additionally the module creates two different figures for the reliability section of the
documentation.
"""
import matplotlib.pyplot as plt
import statsmodels.api as sm
import pandas as pd
import numpy as np
import json

from linearmodels.iv import IV2SLS
from grmpy.simulate.simulate_auxiliary import simulate_unobservables
from grmpy.test.random_init import print_dict
from grmpy.simulate.simulate import simulate
from grmpy.estimate.estimate import fit
from grmpy.read.read import read


def create_data(file):
    """This function creates the a data set based on the results from Caineiro 2011."""
    # Read in initialization file and the data set
    init_dict = read(file)
    df = pd.read_pickle(init_dict['SIMULATION']['source'] + '.grmpy.pkl')

    # Distribute information
    indicator, dep = init_dict['ESTIMATION']['indicator'], init_dict['ESTIMATION']['dependent']
    label_out = [init_dict['varnames'][j - 1] for j in init_dict['TREATED']['order']]
    label_choice = [init_dict['varnames'][j - 1] for j in init_dict['CHOICE']['order']]
    seed = init_dict['SIMULATION']['seed']

    # Set random seed to ensure recomputabiltiy
    np.random.seed(seed)

    # Simulate unobservables
    U, V = simulate_unobservables(init_dict)

    print(len(U[:, 0]))

    df['U1'], df['U0'], df['V'] = U[:, 0], U[:, 1], V

    # Simulate choice and output
    df[dep + '1'] = np.dot(df[label_out], init_dict['TREATED']['all']) + df['U1']
    df[dep + '0'] = np.dot(df[label_out], init_dict['UNTREATED']['all']) + df['U0']
    df[indicator] = np.array(
        np.dot(df[label_choice], init_dict['CHOICE']['all']) - df['V'] > 0).astype(int)
    df[dep] = df[indicator] * df[dep + '1'] + (1 - df[indicator]) * df[dep + '0']

    # Save the data
    df.to_pickle(init_dict['SIMULATION']['source'] + '.grmpy.pkl')

    return df


def update_correlation_structure(model_dict, rho):
    """This function takes a valid model specification and updates the correlation structure
    among the unobservables."""

    # We first extract the baseline information from the model dictionary.
    sd_v = model_dict['DIST']['all'][-1]
    sd_u1 = model_dict['DIST']['all'][0]

    # Now we construct the implied covariance, which is relevant for the initialization file.
    cov1v =  rho * sd_v * sd_u1

    model_dict['DIST']['all'][2] = cov1v

    # We print out the specification to an initialization file with the name mc_init.grmpy.ini.
    for key_ in ['TREATED', 'UNTREATED', 'CHOICE']:
        x = [model_dict['varnames'][j - 1] for j in model_dict[key_]['order']]
        model_dict[key_]['order'] = x
    print_dict(model_dict, 'mc')


def get_effect_grmpy(file):
    """This function simply returns the ATE of the data set."""
    dict_ = read('test.grmpy.ini')
    df = pd.read_pickle('mc.grmpy.pkl')
    beta_diff = dict_['TREATED']['all'] - dict_['UNTREATED']['all']
    covars = [dict_['varnames'][j - 1] for j in dict_['TREATED']['order']]
    ATE = np.dot(np.mean(df[covars]), beta_diff)

    return ATE


def monte_carlo(file, grid_points):
    """This function estimates the ATE for a sample with different correlation structures between U1
     and V. Two different strategies for (OLS,LATE) are implemented.
     """

    # Define a dictionary with a key for each estimation strategy
    effects = {}
    for key_ in ['grmpy', 'ols', 'true', 'random', 'rho', 'iv', 'means']:
        effects[key_] = []

    # Loop over different correlations between V and U_1
    for rho in np.linspace(0.00, 0.99, grid_points):
        effects['rho'] += [rho]
        # Readjust the initialization file values to add correlation
        model_spec = read(file)
        X = [model_spec['varnames'][j - 1] for j in model_spec['TREATED']['order']]
        update_correlation_structure(model_spec, rho)
        model_spec = read('mc.grmpy.ini')

        # Simulate a Data set and specify exogeneous and endogeneous variables
        df_mc = create_data('mc.grmpy.ini')
        endog, exog, exog_ols = df_mc['wage'], df_mc[X], df_mc[['state'] + X]

        instr = [model_spec['varnames'][j - 1] for j in model_spec['CHOICE']['order']]
        instr = [i for i in instr if i != 'const']
        # Calculate true average treatment effect
        ATE = np.mean(df_mc['wage1'] - df_mc['wage0'])
        effects['true'] += [ATE]

        # Estimate  via grmpy
        rslt = fit('test.grmpy.ini')
        beta_diff = rslt['TREATED']['all'] - rslt['UNTREATED']['all']
        stat = np.dot(np.mean(exog), beta_diff)

        effects['grmpy'] += [stat]

        # Estimate via OLS
        ols = sm.OLS(endog, exog_ols).fit()
        stat = ols.params[0]
        effects['ols'] += [stat]

        # Estimate via 2SLS
        iv = IV2SLS(endog, exog, df_mc['state'], df_mc[instr]).fit()
        stat = iv.params['state']
        effects['iv'] += [stat]

        # Estimate via random
        random = np.mean(df_mc[df_mc.state==1]['wage']) - np.mean(df_mc[df_mc.state==0]['wage'])
        stat = random
        effects['random'] += [stat]

        # outcomes
        stat = [[np.mean(df_mc[df_mc.state==1]['wage']),df_mc[df_mc.state==1].shape[0]],[np.mean(df_mc[df_mc.state==0]['wage']),df_mc[df_mc.state==0].shape[0]]]
        effects['means'] += stat

        print(effects)

    return effects


def create_plots(effects, true):
    """The function creates the figures that illustrates the behavior of each estimator of the ATE
    when the correlation structure changes from 0 to 1."""

    fig, ax = plt.subplots(2, 2,figsize=(10,10))

    grid = np.linspace(0.00, 0.99, len(effects['ols']))

    fig.suptitle("Monte Carlo Results", fontsize=16)


    # Determine the title for each strategy plot
    for plot_num1 in [0,1]:
        for plot_num2 in [0, 1]:
            plot_num = [plot_num1, plot_num2]

            true_ = np.tile(true, len(effects['ols']))

            l1, = ax[plot_num1, plot_num2].plot(grid, true_, label="True")
            ax[plot_num1, plot_num2].set_xlim(0, 1)
            ax[plot_num1, plot_num2].set_ylim(0.35, 0.6)
            ax[plot_num1, plot_num2].set_ylabel(r"$B^{ATE}$")
            ax[plot_num1, plot_num2].set_xlabel(r"$\rho_{U_1, V}$")

            ax[plot_num1, plot_num2].yaxis.get_major_ticks()[0].set_visible(False)


            if plot_num == [0, 0]:
                color = 'orange'
                strategy = 'random'
                label = '$E[Y|D=1] - E[Y|D=0]$'
                title = 'Naive comparison'
                l2, = ax[plot_num1, plot_num2].plot(grid, effects[strategy], label=title,
                                                    color=color)

            elif plot_num == [0, 1]:
                color = 'orange'
                label = 'OLS'
                strategy = 'ols'
                title = 'Ordinary Least Squares'
                l2, = ax[plot_num1, plot_num2].plot(grid, effects[strategy], label=title,
                                                    color=color)

            elif plot_num == [1, 0]:
                color = 'orange'
                strategy = 'iv'
                label = 'IV'
                title = 'Instrumental Variables'
                l2, = ax[plot_num1, plot_num2].plot(grid, effects[strategy], label=title,
                                                    color=color)

            elif plot_num == [1, 1]:
                color = 'orange'
                strategy = 'grmpy'
                label = 'grmpy'
                title = 'grmpy'
                l2, = ax[plot_num1, plot_num2].plot(grid, effects[strategy], label=title,
                                                    color=color)
            ax[plot_num1, plot_num2].title.set_text(title)

            ax[plot_num1, plot_num2].legend([l1, l2], ['True', '{}'.format(label)], prop={'size': 8})
    file_name = 'average-effect-estimation.png'.format(strategy)
    plt.savefig(file_name)


if __name__ == '__main__':
    simulate('test.grmpy.ini')

    ATE = get_effect_grmpy('test.grmpy.ini')

    x = monte_carlo('test.grmpy.ini', 10)
    create_plots(x, ATE)





"""
The module provides auxiliary functions for the estimation process.
"""

import numpy as np
import pandas as pd
import statsmodels.api as sm
from numpy.linalg import LinAlgError
from random import randint
from scipy.optimize import minimize
from scipy.stats import norm, t
from statsmodels.tools.numdiff import approx_fprime_cs
from statsmodels.tools.sm_exceptions import PerfectSeparationError

from grmpy.check.check import UserError, check_start_values
from grmpy.estimate.estimate_output import print_logfile


def par_fit(dict_, data):
    """This function runs the parametric estimation of the marginal treatment effect.

    Parameters
    ----------
    dict_: dict
        Estimation dictionary. Returned by grmpy.read(init_file)).
    data: pandas.DataFrame
        Data set to perform the estimation on. Specified
        under dict_["ESTIMATION"]["file"].

    Returns
    ------
    rslt: dict
        Result dictionary containing
        - opt_info: dict
            Dictionary that contains information regarding the conducted optimization
        - opt_rslt: pandas.DataFrame
            Dataframe that provides the results of the optimization process
        - mte
        - mte_x
        - mte_u
        - mte_min
        - mte_max
        - X
        - b1
        - b0
    """

    # process the data set
    D, X1, X0, Z1, Z0, Y1, Y0 = process_data(data, dict_)

    # process optimization options
    opt_dict, method, grad_opt, start_option, print_output, seed_ = process_inputs(
        dict_
    )
    num_treated = X1.shape[1]
    num_untreated = num_treated + X0.shape[1]
    # set seed
    np.random.seed(seed_)

    # set up rslt dataframe object
    rslt_cont = create_rslt_df(dict_)

    # determine start values
    x0 = start_values(dict_, D, X1, X0, Z1, Z0, Y1, Y0, start_option)
    dict_["AUX"]["criteria"] = calculate_criteria(x0, X1, X0, Z1, Z0, Y1, Y0)
    rslt_cont["start_values"] = backward_transformation(x0)
    bfgs_dict = {"parameter": {}, "crit": {}, "grad": {}}
    opt_rslt = minimize(
        minimizing_interface,
        x0,
        args=(X1, X0, Z1, Z0, Y1, Y0, num_treated, num_untreated, bfgs_dict, grad_opt),
        method=method,
        options=opt_dict,
        jac=grad_opt,
    )
    rslt = adjust_output(
        opt_rslt,
        dict_,
        rslt_cont,
        x0,
        method,
        start_option,
        X1,
        X0,
        Z1,
        Z0,
        Y1,
        Y0,
        bfgs_dict,
    )
    # Print Output files
    print_logfile(dict_, rslt, print_output)

    quantiles, cov, X, b1_b0, b1, b0 = prepare_mte_calc(rslt["opt_rslt"], data)
    print(dict_["TREATED"]["order"], dict_["TREATED"]["order"])
    print(b1_b0)
    mte_x = np.dot(X, b1_b0)

    mte_u = (cov[2, 0] - cov[2, 1]) * norm.ppf(quantiles)

    # Put the MTE together
    mte = mte_x.mean(axis=0) + mte_u

    # Account for variation in X
    mte_min = np.min(mte_x) + mte_u
    mte_max = np.max(mte_x) + mte_u

    rslt.update(
        {
            "quantiles": quantiles,
            "mte": mte,
            "mte_x": mte_x,
            "mte_u": mte_u,
            "mte_min": mte_min,
            "mte_max": mte_max,
            "X": X,
            "b1": b1,
            "b0": b0,
        }
    )

    return rslt


def process_data(data, dict_):
    """This function process the data for the optimization process and returns the
    different arrays for the upcoming optimization.

    Parameters
    ----------
    data: pandas.DataFrame
        Data set to perform the estimation on. Specified
        under dict_["ESTIMATION"]["file"].

    dict_: dict
        Estimation dictionary. Returned by grmpy.read(init_file)).

    Returns
    ------
    D: numpy.array
        Treatment indicator
    X1: numpy.array
        Outcome related regressors of the treated individuals
    X0: numpy.array
        Outcome related regressors of the untreated individuals
    Z1: numpy.array
        Choice related regressors of the treated individuals
    Z0: numpy.array
        Choice related regressors of the untreated individuals
    Y1: numpy.array
        Outcomes of the treated individuals
    Y0: numpy.array
        Outcomes of the untreated individuals
    """
    indicator = dict_["ESTIMATION"]["indicator"]
    outcome = dict_["ESTIMATION"]["dependent"]
    D = data[indicator].values

    data1 = data[data[indicator] == 1]
    data2 = data[data[indicator] == 0]

    X1 = data1[dict_["TREATED"]["order"]].values
    X0 = data2[dict_["UNTREATED"]["order"]].values
    Z1 = data1[dict_["CHOICE"]["order"]].values
    Z0 = data2[dict_["CHOICE"]["order"]].values

    Y1 = data1[outcome].values
    Y0 = data2[outcome].values

    return D, X1, X0, Z1, Z0, Y1, Y0


def process_inputs(dict_):
    """This function processes the specifications related to the optimzation routine.

    Parameters
    ----------
    dict_: dict_
        Estimation dictionary. Returned by grmpy.read(init_file)).

    Returns
    ------
    opt_dict: dict
        Solver options
    method: str
        Algorithm that is used for the minimization
    grad_opt: bool
        Boolean that determines whether the chosen algorithm is derivative based
    start_option: str
        Denotes which start value routine should be used. Options are
        either "init" or "auto".
    print_output: bool
        If True the estimation output is printed
    seed_: int
        Seed value for drawing the random start values for the rho1 and rho0
    """
    try:
        method = dict_["ESTIMATION"]["optimizer"]
    except KeyError:
        method = "BFGS"

    grad_opt = method == "BFGS"

    try:
        opt_dict = dict_["SCIPY-" + method]
    except KeyError:
        opt_dict = {}

    try:
        start_option = dict_["ESTIMATION"]["start"]
    except KeyError:
        start_option = "auto"

    try:
        opt_dict["maxiter"] = dict_["ESTIMATION"]["maxiter"]
        if opt_dict["maxiter"] == 0:
            start_option = "init"
    except KeyError:
        pass

    try:
        print_output = dict_["ESTIMATION"]["print_output"]
    except KeyError:
        print_output = True

    try:
        seed_ = dict_["SIMULATION"]["seed"]
    except KeyError:
        seed_ = randint(0, 9999)

    return opt_dict, method, grad_opt, start_option, print_output, seed_


def create_rslt_df(dict_):
    """This function creates the pandas dataframe container in which the estimation
    rslts are stored.

    Parameters
    ----------
    dict_: dict
        Estimation dictionary. Returned by grmpy.read(init_file)).

    Returns
    ------
    rslt_cont: pandas.DataFrame
        Container for stroing the upcoming estimation results.
    """

    index = []
    # set up multiindex
    for section in ["TREATED", "UNTREATED", "CHOICE"]:
        index += [(section, i) for i in dict_[section]["order"]]
    for subsection in ["sigma1", "rho1", "sigma0", "rho0"]:
        index += [("DIST", subsection)]

    column_names = [
        "params",
        "start_values",
        "std",
        "t_values",
        "p_values",
        "conf_int_low",
        "conf_int_up",
    ]

    return pd.DataFrame(
        index=pd.MultiIndex.from_tuples(index, names=["section", "name"]),
        columns=column_names,
    )


def start_values(dict_, D, X1, X0, Z1, Z0, Y1, Y0, start_option):
    """The function selects the start values for the minimization process. If option is
    set to init the function returns the values that are specified in the initialization
    file. Otherwise the function conducts a Probit estimation for determining the choice
    related parameters as well as two OLS estimations for the outcome related parameters
    associated with the different treatment states. In this case the sigma values are set
    to the sum of residual squares of the particular OLS regression, whereas the rho
    values are drawn randomly. Finally the sigma and rho values are converted by applying
    a method based on Lokshin and Sajaia (2004) independent on the chosen start value
    option.

    Parameters
    ----------
    dict_: dict
        Estimation dictionary. Returned by grmpy.read(init_file)).
    D: numpy.array
        Treatment indicator
    X1: numpy.array
        Outcome related regressors of the treated individuals.
    X0: numpy.array
        Outcome related regressors of the untreated individuals.
    Z1: numpy.array
        Choice related regressors of the treated individuals.
    Z0: numpy.array
        Choice related regressors of the untreated individuals.
    Y1: numpy.array
        Outcomes of the treated individuals.
    Y0: numpy.array
        Outcomes of the untreated individuals.
    start_option: str
        Denotes which start value routine should be used. Options are
        either "init" or "auto".

    Returns
    ------
    x0: numpy.array
        Start values for the estimation routine
    """
    if not isinstance(dict_, dict):
        msg = (
            "The input object ({})for specifing the start values isn`t a "
            "dictionary.".format(dict_)
        )
        raise UserError(msg)

    if start_option == "init":
        # Set coefficients equal the true init file values
        rho1 = dict_["DIST"]["params"][2] / dict_["DIST"]["params"][0]
        rho0 = dict_["DIST"]["params"][4] / dict_["DIST"]["params"][3]
        dist = [dict_["DIST"]["params"][0], rho1, dict_["DIST"]["params"][3], rho0]
        x0 = np.concatenate((dict_["AUX"]["init_values"][:-6], dist))
    elif start_option == "auto":
        try:
            if D.shape[0] == sum(D):
                raise PerfectSeparationError
            # Estimate beta1 and beta0:
            beta = []
            sd_ = []

            for data_out in [(Y1, X1), (Y0, X0)]:
                ols_results = sm.OLS(data_out[0], data_out[1]).fit()
                beta += [ols_results.params]
                sd = np.sqrt(ols_results.scale)
                rho = np.random.uniform(-sd, sd, 1) / sd
                sd_ += [sd, rho[0]]

            # Estimate gamma via Probit
            Z = np.vstack((Z0, Z1))
            probitRslt = sm.Probit(np.sort(D), Z).fit(disp=0)
            gamma = probitRslt.params
            # Adjust estimated cost-benefit shifter and intercept coefficients
            # Arrange starting values
            x0 = np.concatenate((beta[0], beta[1], gamma, sd_))
            check_start_values(x0)

        except (PerfectSeparationError, ValueError, UserError):
            msg = (
                "The estimation process wasn`t able to provide automatic"
                " start values due to perfect seperation. \n"
                " The intialization specifications are used as start "
                "values during the further process."
            )

            # Set coefficients equal the true init file values
            rho1 = dict_["DIST"]["params"][2] / dict_["DIST"]["params"][0]
            rho0 = dict_["DIST"]["params"][4] / dict_["DIST"]["params"][3]
            dist = [dict_["DIST"]["params"][0], rho1, dict_["DIST"]["params"][3], rho0]
            x0 = np.concatenate((dict_["AUX"]["init_values"][:-6], dist))
            dict_["ESTIMATION"]["warning"] = msg
            start_option = "init"

    x0 = start_value_adjustment(x0)
    return np.array(x0)


def start_value_adjustment(x):
    """This function conducts an approach by Lokshin and Sajaia (2004) and takes the
    logarithm of the sigma values as well as the inverse hyperbolic tangent of the speci-
    fied start values for the rhovalues. The transformation will be inversed right within
    the minimization interface function. Through this we ensure that the estimated values
    for sigma are always larger than 0 and that the rho values are bounded between -1 and
    1.

    Parameters
    ----------
    x: numpy.array
        Start values for the estimation routine

    Returns
    ------
    x: numpy.array
        Transformed start values for the estimation routine
    """

    # transform the distributional characteristics s.t. r = log((1-rho)/(1+rho))/2
    x[-4:] = [
        np.log(x[-4]),
        np.log((1 + x[-3]) / (1 - x[-3])) / 2,
        np.log(x[-2]),
        np.log((1 + x[-1]) / (1 - x[-1])) / 2,
    ]

    return x


def backward_transformation(x_trans, bfgs_dict=None):
    """This function reverses the transformation of the sigma and rho values.

    Parameters
    ----------
    x_trans: numpy.array
        Transformed parameter values
    bfgs_dict: dict
        Dictionary that logs the different parameterizations that are
        evaluated during the minimization.
    Returns
    ------
    x_rev: numpy.array
        Reversed parameter values
    """
    x_rev = x_trans.copy()
    x_rev[-4:] = [
        np.exp(x_rev[-4]),
        (np.exp(2 * x_rev[-3]) - 1) / (np.exp(2 * x_rev[-3]) + 1),
        np.exp(x_rev[-2]),
        (np.exp(2 * x_rev[-1]) - 1) / (np.exp(2 * x_rev[-1]) + 1),
    ]
    if bfgs_dict is not None:
        bfgs_dict["parameter"][str(len(bfgs_dict["parameter"]))] = x_rev
    return x_rev


def log_likelihood(
    x0,
    X1,
    X0,
    Z1,
    Z0,
    Y1,
    Y0,
    num_treated,
    num_untreated,
    bfgs_dict=None,
    grad_opt=True,
):
    """This is the the log-likelihood function of our minimization problem.

    Parameters
    ----------
    x0: numpy.array
        Parameter values
    X1: numpy.array
        Outcome related regressors of the treated individuals
    X0: numpy.array
        Outcome related regressors of the untreated individuals
    Z1: numpy.array
        Choice related regressors of the treated individuals
    Z0: numpy.array
        Choice related regressors of the untreated individuals
    Y1: numpy.array
        Outcomes of the treated individuals
    Y0: numpy.array
        Outcomes of the untreated individuals
    bfgs_dict: dict
        Dictionary that logs the different parameterizations that are
        evaluated during the minimization.
    grad_opt: bool
        If True, the function returns not only the likelihood value
        but also the gradient

    Returns
    ------
    likl: float
        Negative log-likelihood value
    llh_grad: numpy.array
        Gradient of the minimization interface, only returned if grad_opt==True
    """

    # assign parameter values
    beta1, beta0, gamma = (
        x0[:num_treated],
        x0[num_treated:num_untreated],
        x0[num_untreated:-4],
    )
    sd1, sd0, rho1v, rho0v = x0[-4], x0[-2], x0[-3], x0[-1]

    nu1 = (Y1 - np.dot(beta1, X1.T)) / sd1
    lambda1 = (np.dot(gamma, Z1.T) - rho1v * nu1) / (np.sqrt(1 - rho1v ** 2))

    nu0 = (Y0 - np.dot(beta0, X0.T)) / sd0
    lambda0 = (np.dot(gamma, Z0.T) - rho0v * nu0) / (np.sqrt(1 - rho0v ** 2))

    treated = (1 / sd1) * norm.pdf(nu1) * norm.cdf(lambda1)
    untreated = (1 / sd0) * norm.pdf(nu0) * (1 - norm.cdf(lambda0))

    likl = -np.mean(np.log(np.append(treated, untreated)))

    if bfgs_dict is not None:
        bfgs_dict["crit"][str(len(bfgs_dict["crit"]))] = likl

    if grad_opt is True:
        llh_grad = gradient(
            X1, X0, Z1, Z0, nu1, nu0, lambda1, lambda0, gamma, sd1, sd0, rho1v, rho0v
        )
        return likl, llh_grad
    else:
        return likl


def calculate_criteria(x0, X1, X0, Z1, Z0, Y1, Y0):
    """The function computes the criterion function value for a given parameter specifi-
    cation.

    Parameters
    ----------
    x0: numpy.array
        Parameter values
    X1: numpy.array
        Outcome related regressors of the treated individuals
    X0: numpy.array
        Outcome related regressors of the untreated individuals
    Z1: numpy.array
        Choice related regressors of the treated individuals
    Z0: numpy.array
        Choice related regressors of the untreated individuals
    Y1: numpy.array
        Outcomes of the treated individuals
    Y0: numpy.array
        Outcomes of the untreated individuals

    Returns
    ------
    crit_value: float
        criterion function value
    """
    x = backward_transformation(x0)
    num_treated = X1.shape[1]
    num_untreated = num_treated + X0.shape[1]

    return log_likelihood(
        x, X1, X0, Z1, Z0, Y1, Y0, num_treated, num_untreated, None, False
    )


def minimizing_interface(
    x0, X1, X0, Z1, Z0, Y1, Y0, num_treated, num_untreated, bfgs_dict, grad_opt
):
    """This function is the objective for the minimization routine. It transforms the
    provided parameters according to the and returns the associated log-likelihood value.

    Parameters
    ----------
    x0: numpy.array
        Parameter values
    X1: numpy.array
        Outcome related regressors of the treated individuals
    X0: numpy.array
        Outcome related regressors of the untreated individuals
    Z1: numpy.array
        Choice related regressors of the treated individuals
    Z0: numpy.array
        Choice related regressors of the untreated individuals
    Y1: numpy.array
        Outcomes of the treated individuals
    Y0: numpy.array
        Outcomes of the untreated individuals
    num_treated: float
        number of regressors of the outcome equation for the treated individuals
    num_untreated: numpy.array
        number of regressors of the outcome equation for the treated
        and untreated individuals
    bfgs_dict: dict
        Dictionary that logs the different parameterizations that are
        evaluated during the minimization.
    grad_opt: bool
        If True, the function returns not only the likelihood value
        but also the gradient
    Returns
    ------
    likl: float
        Negative log-likelihood value
    llh_grad: numpy.array
        Jacobian of the minimization interface, only returned if grad_opt==True
    """

    # transform input parameter
    x0 = backward_transformation(x0, bfgs_dict)

    # Calculate likelihood
    return log_likelihood(
        x0, X1, X0, Z1, Z0, Y1, Y0, num_treated, num_untreated, bfgs_dict, grad_opt
    )


def adjust_output(
    opt_rslt,
    dict_,
    rslt_cont,
    start_values,
    method,
    start_option,
    X1,
    X0,
    Z1,
    Z0,
    Y1,
    Y0,
    bfgs_dict=None,
):
    """The function adds different information of the minimization process to the
    estimation output.
    """

    rslt = {
        "opt_info": {
            "optimizer": method,
            "start": start_option,
            "indicator": dict_["ESTIMATION"]["indicator"],
            "dependent": dict_["ESTIMATION"]["dependent"],
            "observations": Y1.shape[0] + Y0.shape[0],
        }
    }
    # Adjust output if
    if opt_rslt["nit"] == 0:
        x = backward_transformation(opt_rslt["x"])
        rslt["opt_info"]["success"], rslt["opt_info"]["status"] = False, 2
        (
            rslt["opt_info"]["message"],
            rslt["opt_info"]["nit"],
            rslt["opt_info"]["crit"],
        ) = ("---", 0, dict_["AUX"]["criteria"])
        rslt["opt_info"]["warning"] = ["---"]

    else:
        # Check if the algorithm has returned the values with the lowest criterium
        # function value
        check, flag = check_rslt_parameters(
            opt_rslt["x"], X1, X0, Z1, Z0, Y1, Y0, bfgs_dict
        )
        # Adjust values if necessary
        if check:
            x, crit, warning = process_output(dict_, bfgs_dict, opt_rslt["x"], flag)
            rslt["opt_info"]["crit"] = crit
            rslt["opt_info"]["warning"] = [warning]

        else:
            x = backward_transformation(opt_rslt["x"])
            rslt["opt_info"]["crit"] = opt_rslt["fun"]
            rslt["opt_info"]["warning"] = ["---"]

        (
            rslt["opt_info"]["success"],
            rslt["opt_info"]["status"],
            rslt["opt_info"]["message"],
            rslt["opt_info"]["nit"],
        ) = map(opt_rslt.get, ["success", "status", "message", "nit"])

    # Adjust Result dict
    rslt_cont["params"] = x

    (
        rslt_cont["std"],
        rslt["hessian_inv"],
        rslt_cont["conf_int_low"],
        rslt_cont["conf_int_up"],
        rslt_cont["p_values"],
        rslt_cont["t_values"],
        warning_se,
    ) = calculate_se(x, dict_, X1, X0, Z1, Z0, Y1, Y0)

    rslt["opt_rslt"] = rslt_cont

    if warning_se is not None:
        rslt["opt_info"]["warning"] += warning_se
    return rslt


def check_rslt_parameters(x0, X1, X0, Z1, Z0, Y1, Y0, bfgs_dict):
    """This function checks if the algorithm has not touched a parameterization during
    the optimization process that leads to a lower criterion function value than the one
    that the minimization routine returned.

    Parameters
    ----------
    x0: numpy.array
        Parameter values
    X1: numpy.array
        Outcome related regressors of the treated individuals
    X0: numpy.array
        Outcome related regressors of the untreated individuals
    Z1: numpy.array
        Choice related regressors of the treated individuals
    Z0: numpy.array
        Choice related regressors of the untreated individuals
    Y1: numpy.array
        Outcomes of the treated individuals
    Y0: numpy.array
        Outcomes of the untreated individuals
    bfgs_dict: dict
        Dictionary that logs the different parameterizations that are
        evaluated during the minimization.

    Returns
    ------
    check: bool
        True if the bfgs_dict contains a parameterization that leads to a smaller
        log-likelihood value or if the array contains nan/nonfinite values
    flag: str
        Flag that indicates whether the array contained nan/nonfinite values or if there
        the algorithm reached a parameterization that leads to smaller likelihood values
        than the one that the algorithm converged to.
    """

    crit = calculate_criteria(x0, X1, X0, Z1, Z0, Y1, Y0)
    x = min(bfgs_dict["crit"], key=bfgs_dict["crit"].get)
    if False in np.isfinite(x0).tolist():
        check, flag = True, "notfinite"

    elif bfgs_dict["crit"][str(x)] < crit:
        check, flag = True, "adjustment"

    else:
        check, flag = False, None
    return check, flag


def process_output(init_dict, bfgs_dict, x0, flag):
    """The function checks if the criteria function value is smaller for the
    optimization output as for the start values.

    Parameters
    ----------
    x0: numpy.array
        Parameter values
    X1: numpy.array
        Outcome related regressors of the treated individuals
    X0: numpy.array
        Outcome related regressors of the untreated individuals
    Z1: numpy.array
        Choice related regressors of the treated individuals
    Z0: numpy.array
        Choice related regressors of the untreated individuals
    Y1: numpy.array
        Outcomes of the treated individuals
    Y0: numpy.array
        Outcomes of the untreated individuals
    bfgs_dict: dict
        Dictionary that logs the different parameterizations that are
        evaluated during the minimization.

    Returns
    ------
    check: bool
        True if the bfgs_dict contains a parameterization that leads to a smaller
        log-likelihood value or if the array contains nan/nonfinite values
    flag: str
        Flag that indicates whether the array contained nan/nonfinite values or if there
        the algorithm reached a parameterization that leads to smaller likelihood values
        than the one that the algorithm converged to.
    """

    x = min(bfgs_dict["crit"], key=bfgs_dict["crit"].get)

    if flag == "adjustment":
        if bfgs_dict["crit"][str(x)] < init_dict["AUX"]["criteria"]:
            x0 = bfgs_dict["parameter"][str(x)].tolist()
            crit = bfgs_dict["crit"][str(x)]
            warning = (
                "The optimization algorithm has failed to provide the "
                "parametrization that leads to the minimal criterion function "
                "value.The estimation output is automatically adjusted and "
                "provides the parameterization with the smallest criterion "
                "function value that was reached during the optimization."
            )
        else:
            x0 = x0
            crit = bfgs_dict["crit"][str(x)]
            warning = "NONE"

    elif flag == "notfinite":
        x0 = init_dict["AUX"]["starting_values"]
        crit = init_dict["AUX"]["criteria"]
        warning = (
            "The optimization process is not able to provide finite values. This is "
            "probably due to perfect separation."
        )
    else:
        crit = x
    return x0, crit, warning


def calculate_se(x, maxiter, X1, X0, Z1, Z0, Y1, Y0):
    """This function computes the standard errors of the parameters by approximating the
    Jacobian of the gradient function. Based on that it computes the confidence

    Parameters
    ----------
    x: numpy.array
        Parameter values
    maxiter: float
        maximum number of iterations
    X1: numpy.array
        Outcome related regressors of the treated individuals
    X0: numpy.array
        Outcome related regressors of the untreated individuals
    Z1: numpy.array
        Choice related regressors of the treated individuals
    Z0: numpy.array
        Choice related regressors of the untreated individuals
    Y1: numpy.array
        Outcomes of the treated individuals
    Y0: numpy.array
        Outcomes of the untreated individuals

    Returns
    ------
    se: numpy.array
        Standard errors of the parameters
    hess_inv: numpy.array
        Inverse hessian matrix evaluated at the parameter vector
    conf_interval: numpy.array
        Confidence intervals of the parameters
    p_values: numpy.array
        p-values of the parameters
    t_values: numpy.array
        t-values of the parameters
    warning: str
        Warning message if the approximated hessian matrix is not invertible

    """
    num_ind = Y1.shape[0] + Y0.shape[0]
    x0 = x.copy()
    warning = None

    if maxiter == 0:
        se = [np.nan] * len(x0)
        hess_inv = np.full((len(x0), len(x0)), np.nan)
        conf_interval = [[np.nan, np.nan]] * len(x0)
        p_values, t_values = len(x0) * [np.nan], len(x0) * [np.nan]
    else:
        norm_value = norm.ppf(0.975)
        # Calculate the hessian matrix, check if it is p
        hess = approx_fprime_cs(x0, gradient_hessian, args=(X1, X0, Z1, Z0, Y1, Y0))
        try:
            hess_inv = np.linalg.inv(hess)
            se = np.sqrt(np.diag(hess_inv) / num_ind)
            aux = norm_value * se
            hess_inv = hess_inv
            conf_interval = np.vstack((np.subtract(x0, aux), np.add(x0, aux))).T
            t_values = np.divide(x0, se)
            p_values = 2 * (1 - t.cdf(np.abs(t_values), df=num_ind - len(x0)))

        except LinAlgError:
            se = np.full(len(x0), np.nan)
            hess_inv = np.full((len(x0), len(x0)), np.nan)
            conf_interval = np.full((len(x0), 2), np.nan)
            t_values = np.full(len(x0), np.nan)
            p_values = np.full(len(x0), np.nan)

        # Check if standard errors are defined, if not add warning message

        if False in np.isfinite(se):
            warning = [
                "The estimation process was not able to provide standard errors for"
                " the estimation results, because the approximation of the hessian "
                "matrix leads to a singular Matrix."
            ]

    return (
        se,
        hess_inv,
        conf_interval[:, 0],
        conf_interval[:, 1],
        p_values,
        t_values,
        warning,
    )


def gradient(X1, X0, Z1, Z0, nu1, nu0, lambda1, lambda0, gamma, sd1, sd0, rho1v, rho0v):
    """This function returns the gradient of our minimization interface.

    Parameters
    ----------
    X1: numpy.array
        Outcome related regressors of the treated individuals
    X0: numpy.array
        Outcome related regressors of the untreated individuals
    Z1: numpy.array
        Choice related regressors of the treated individuals
    Z0: numpy.array
        Choice related regressors of the untreated individuals
    Y1: numpy.array
        Outcomes of the treated individuals
    Y0: numpy.array
        Outcomes of the untreated individuals
    nu1: numpy.array
        residual of the outcome equation of the treated individuals divided by sigma 1
    nu0: numpy.array
        residual of the outcome equation of the untreated individuals divided by sigma 1
    lambda1: numpy.array
        (gamma * Z / rho1 * nu1) / sqrt(1 -  rho1 ** 2)
    lambda0: numpy.array
        (gamma * Z / rho0 * nu0) / sqrt(1 -  rho0 ** 2)
    gamma: numpy.array
        Choice related parameters
    sd1: float
        Sigma 1
    sd0: float
        Sigma 0
    rho1v: float
        rho1
    rho0v: float
        rho0

    Returns
    ------
    grad: numpy.array
        Gradient of the minimization interface
    """
    n_obs = X1.shape[0] + X0.shape[0]

    # compute gradient coef for beta 1

    grad_beta1 = np.sum(
        np.einsum(
            "ij, i ->ij",
            X1,
            -(norm.pdf(lambda1) / norm.cdf(lambda1))
            * (rho1v / (np.sqrt(1 - rho1v ** 2) * sd1))
            - nu1 / sd1,
        ),
        0,
    )

    # compute coef for beta 0
    grad_beta0 = np.sum(
        np.einsum(
            "ij, i ->ij",
            X0,
            norm.pdf(lambda0)
            / (1 - norm.cdf(lambda0))
            * (rho0v / (np.sqrt(1 - rho0v ** 2) * sd0))
            - nu0 / sd0,
        ),
        0,
    )
    grad_sd1 = np.sum(
        sd1
        * (
            +1 / sd1
            - (norm.pdf(lambda1) / norm.cdf(lambda1))
            * (rho1v * nu1 / (np.sqrt(1 - rho1v ** 2) * sd1))
            - nu1 ** 2 / sd1
        ),
        keepdims=True,
    )
    grad_sd0 = np.sum(
        sd0
        * (
            +1 / sd0
            + (norm.pdf(lambda0) / (1 - norm.cdf(lambda0)))
            * (rho0v * nu0 / (np.sqrt(1 - rho0v ** 2) * sd0))
            - nu0 ** 2 / sd0
        ),
        keepdims=True,
    )
    grad_rho1v = np.sum(
        (
            -(norm.pdf(lambda1) / norm.cdf(lambda1))
            * ((np.dot(gamma, Z1.T) * rho1v) - nu1)
            / (1 - rho1v ** 2) ** (1 / 2)
        ),
        keepdims=True,
    )

    grad_rho0v = np.sum(
        (
            (norm.pdf(lambda0) / (1 - norm.cdf(lambda0)))
            * ((np.dot(gamma, Z0.T) * rho0v) - nu0)
            / (1 - rho0v ** 2) ** (1 / 2)
        ),
        keepdims=True,
    )

    grad_gamma = +sum(
        np.einsum(
            "ij, i ->ij",
            Z1,
            (norm.pdf(lambda1) / norm.cdf(lambda1)) * 1 / np.sqrt(1 - rho1v ** 2),
        )
    ) - sum(
        np.einsum(
            "ij, i ->ij",
            Z0,
            (norm.pdf(lambda0) / (1 - norm.cdf(lambda0)))
            * (1 / np.sqrt(1 - rho0v ** 2)),
        )
    )

    grad = np.concatenate(
        (
            grad_beta1,
            grad_beta0,
            -grad_gamma,
            grad_sd1,
            grad_rho1v,
            grad_sd0,
            grad_rho0v,
        )
    )

    return grad / n_obs


def gradient_hessian(x0, X1, X0, Z1, Z0, Y1, Y0):
    """This function computes the gradient of our log-likelihood function at a given
    paramterization x0. The function is used to approximate the hessian matrix for the
    calculation of the standard errors.

    Parameters
    ----------
    x0: numpy.array
        Parameter values
    X1: numpy.array
        Outcome related regressors of the treated individuals
    X0: numpy.array
        Outcome related regressors of the untreated individuals
    Z1: numpy.array
        Choice related regressors of the treated individuals
    Z0: numpy.array
        Choice related regressors of the untreated individuals
    Y1: numpy.array
        Outcomes of the treated individuals
    Y0: numpy.array
        Outcomes of the untreated individuals

    Returns
    ------
    grad: numpy.array
        Gradient of the log-likelihood function.
    """
    num_treated = X1.shape[1]
    num_untreated = num_treated + X0.shape[1]

    beta1, beta0, gamma = (
        x0[:num_treated],
        x0[num_treated:num_untreated],
        x0[num_untreated:-4],
    )
    sd1, sd0, rho1v, rho0v = x0[-4], x0[-2], x0[-3], x0[-1]

    # compute gradient for beta 1

    nu1 = (Y1 - np.dot(beta1, X1.T)) / sd1
    lambda1 = (np.dot(gamma, Z1.T) - rho1v * nu1) / (np.sqrt(1 - rho1v ** 2))

    nu0 = (Y0 - np.dot(beta0, X0.T)) / sd0
    lambda0 = (np.dot(gamma, Z0.T) - rho0v * nu0) / (np.sqrt(1 - rho0v ** 2))

    grad = gradient(
        X1, X0, Z1, Z0, nu1, nu0, lambda1, lambda0, gamma, sd1, sd0, rho1v, rho0v
    )

    multiplier = np.concatenate(
        (
            np.ones(len(grad[:-4])),
            np.array([1 / sd1, 1 / (1 - rho1v ** 2), 1 / sd0, 1 / (1 - rho0v ** 2)]),
        )
    )

    return multiplier * grad


def prepare_mte_calc(opt_rslt, data):
    """This function construct the marginal treatment effect
    given the optimization results.

    Parameters
    ----------
    opt_rslt: pandas.DataFrame
        Dataframe that contains the optimization results.
    data: pandas.DataFrame
        Data set to perform the estimation on.

    Returns
    ------
    quantiles: numpy.array
        Grid on which the marginal treatment effect got evaluated.
    cov: numpy.array
        Covariance matrix based on the estimation results.
    X: numpy.array
        Array object that contains all covariates that affect the treated
        as well as the untreated outcome.
    b1_b0: numpy.array
        Combined array of the parameters for the outcome equations.
    b1: numpy.array
        Estimated beta1 vector.
    b0: numpy.array
        Estimated beta0 vector.
    """

    quantiles = [0.0001] + np.arange(0.01, 1.0, 0.01).tolist() + [0.9999]

    # create a proper covariance matrix from the estimation results
    dist_params = opt_rslt.loc["DIST", "params"]
    cov = np.zeros((3, 3))
    np.fill_diagonal(cov, np.array([dist_params[0], dist_params[2], 1.0]) ** 2)
    cov[2, 0] = dist_params[0] * dist_params[1]
    cov[2, 1] = dist_params[2] * dist_params[3]

    x_treated = data[opt_rslt.loc["TREATED"].index.values].values
    x_untreated = data[opt_rslt.loc["UNTREATED"].index.values].values
    X = np.append(x_treated, -x_untreated, axis=1)

    beta1 = opt_rslt.loc["TREATED"].params.values
    beta0 = opt_rslt.loc["UNTREATED"].params.values

    b1_b0 = np.append(beta1, beta0)

    return quantiles, cov, X, b1_b0, beta1, beta0



"""This module contains the custom exceptions for the package."""


class UserError(Exception):
    """This custom error class provides informative feedback in case of a misspecified
    request by the user.
    """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return f"\n\n         {self.msg}\n\n"



"""The module provides a random dictionary generating process for test purposes."""
import collections
import uuid
from itertools import combinations

import numpy as np
import oyaml as yaml
from scipy.stats import wishart

from grmpy.check.check import UserError


def generate_random_dict(constr=None):
    """The module provides a random dictionary generating process for test purposes."""

    if constr is not None:
        if not isinstance(constr, dict):
            msg = f"{constr} is not a dictionary."
            raise UserError(msg)
    else:
        constr = dict()

    if "DETERMINISTIC" in constr.keys():
        is_deterministic = constr["DETERMINISTIC"]
    else:
        is_deterministic = np.random.random_sample() < 0.1

    if "COMPARISON" in constr.keys():
        comparison = constr["COMPARISON"]
    else:
        comparison = False

    if "STATE_DIFF" in constr.keys():
        state_diff = constr["STATE_DIFF"]
    else:
        state_diff = np.random.random_sample() < 0.5

    if "BINARY" in constr.keys():
        allow_binary = constr["BINARY"]
    else:
        allow_binary = True

    if "OPTIMIZER" in constr.keys():
        optimizer = constr["OPTIMIZER"]
    else:
        optimizer = str(np.random.choice(a=["BFGS", "POWELL"], p=[0.5, 0.5]))

    if "SAME_SIZE" in constr.keys():
        same_size = constr["SAME_SIZE"]
    else:
        same_size = False

    if "IS_ZERO" in constr.keys() and not is_deterministic:
        is_zero = np.random.random_sample() < 0.1 / (1 - 0.1)
    else:
        is_zero = False

    if "OVERLAP" in constr.keys():
        overlap = constr["OVERLAP"]
    else:
        overlap = np.random.random_sample() < 0.5

    if "MAXITER" in constr.keys():
        maxiter = constr["MAXITER"]
    else:
        maxiter = np.random.randint(0, 10000)

    if "AGENTS" in constr.keys():
        agents = constr["AGENTS"]

    else:
        agents = np.random.randint(1, 1000)

    if "START" in constr.keys():
        start = constr["START"]
    else:
        start = str(np.random.choice(a=["init", "auto"]))
    if "SEED" in constr.keys():
        seed = constr["SEED"]
    else:
        seed = int(np.random.randint(1, 10000))

    source = str(uuid.uuid4()).upper().replace("-", "")[0:8]

    # Specify the number of variables/parameters for every section
    treated_num = np.random.randint(1, 10)
    choice_num = np.random.randint(1, 10)

    # Determine if there are different variables that affect the outcome states
    if state_diff:
        untreated_num = treated_num + np.random.randint(1, 10)
        choice_num = untreated_num + choice_num
        num = [
            [1, treated_num],
            [treated_num, untreated_num],
            [untreated_num, choice_num],
        ]

    else:
        untreated_num = treated_num
        choice_num = treated_num + choice_num
        num = [[1, treated_num], [1, treated_num], [treated_num, choice_num]]

    init_dict = {}

    # Coefficients
    for counter, section in enumerate(["TREATED", "UNTREATED", "CHOICE"]):
        init_dict[section] = {}
        init_dict[section]["params"], init_dict[section]["order"] = generate_coeff(
            num[counter], is_zero
        )

    # Specify if there are variables that affect a combination of sections
    init_dict = comb_overlap(init_dict, state_diff, overlap)

    # Specify if some variables are binary
    init_dict["VARTYPES"] = {}
    for variable in set(
        init_dict["TREATED"]["order"]
        + init_dict["UNTREATED"]["order"]
        + init_dict["CHOICE"]["order"]
    ):
        init_dict["VARTYPES"][variable] = "nonbinary"

    if allow_binary is True:
        init_dict = types(init_dict)

    # Simulation parameters
    init_dict["SIMULATION"] = {}
    init_dict["SIMULATION"]["seed"] = seed
    init_dict["SIMULATION"]["agents"] = agents
    init_dict["SIMULATION"]["source"] = source

    # Estimation parameters
    init_dict["ESTIMATION"] = {}
    if same_size:
        init_dict["ESTIMATION"]["agents"] = agents
    else:
        init_dict["ESTIMATION"]["agents"] = np.random.randint(1, 1000)
    init_dict["ESTIMATION"]["file"] = source + ".grmpy.txt"
    init_dict["ESTIMATION"]["optimizer"] = optimizer
    init_dict["ESTIMATION"]["start"] = start
    init_dict["ESTIMATION"]["maxiter"] = maxiter
    init_dict["ESTIMATION"]["dependent"] = "Y"
    init_dict["ESTIMATION"]["indicator"] = "D"
    init_dict["ESTIMATION"]["output_file"] = "est.grmpy.info"
    init_dict["ESTIMATION"]["comparison"] = comparison
    init_dict["ESTIMATION"]["print_output"] = False

    init_dict["SCIPY-BFGS"], init_dict["SCIPY-POWELL"] = {}, {}
    init_dict["SCIPY-BFGS"]["gtol"] = np.random.uniform(1.5e-05, 0.8e-05)
    init_dict["SCIPY-BFGS"]["eps"] = np.random.uniform(
        1.4901161193847655e-08, 1.4901161193847657e-08
    )
    init_dict["SCIPY-POWELL"]["xtol"] = np.random.uniform(0.00009, 0.00011)
    init_dict["SCIPY-POWELL"]["ftol"] = np.random.uniform(0.00009, 0.00011)

    # Variance and covariance parameters
    init_dict["DIST"] = {}
    if not is_deterministic:
        scale_matrix = np.identity(3) * 0.1
        b = wishart.rvs(df=10, scale=scale_matrix, size=1)
        for i in [0, 1, 2]:
            b[i, i] = b[i, i] ** 0.5
    else:
        b = np.zeros((3, 3))
    init_dict["DIST"]["params"] = np.around(
        [float(i) for i in list(b[np.triu_indices(3)])], 4
    ).tolist()

    print_dict(init_dict)

    return init_dict


def generate_coeff(num, is_zero):
    """The function generates random coefficients for creating the random init
    dictionary.
    """

    # Generate a random paramterization and specify the variable order
    if not is_zero:
        params = np.around(
            np.random.normal(0.0, 2.0, [len(range(num[0] - 1, num[1]))]), 4
        ).tolist()
    else:
        params = np.array([0] * num).tolist()

    order = ["X1"] + [f"X{i + 1}" for i in range(num[0], num[1])]

    return params, order


def types(init_dict):
    """This function determines if there are any binary variables. If so the function
    specifies the rate for which the variable is equal to one.
    """

    variables = [i for i in init_dict["VARTYPES"].keys() if i != "X1"]
    for var in variables:
        if np.random.random_sample() < 0.1:
            frac = np.random.uniform(0, 0.8)
            init_dict["VARTYPES"][var] = ["binary", frac]
        else:
            pass

    return init_dict


def comb_overlap(init_dict, state_diff, overlap):
    """This function evaluates which variables affect more than one section."""

    # List all possible overlaps between the different sections and chose a random
    # combination
    if state_diff and overlap:
        cases = [list(i) for i in combinations(list(init_dict.keys()), 2)] + [
            list(init_dict.keys())
        ]
        case = np.random.choice(cases)
        case = [i for i in case if len(init_dict[i]["order"]) > 1]
    elif not state_diff and overlap:
        case = list(init_dict.keys())
        case = [i for i in case if len(init_dict[i]["order"]) > 1]
    else:
        case = []

    # Select a random number of variables that effect the chosen combination of sections
    if len(case) != 0:
        aux_dict = {j: len(init_dict[j]["order"]) for j in case}
        min_key = min(aux_dict, key=aux_dict.get)
        num_overlap = np.random.choice(range(1, aux_dict[min_key]))
        for section in case:
            init_dict[section]["order"][1 : 1 + num_overlap] = init_dict[min_key][
                "order"
            ][1 : num_overlap + 1]

    return init_dict


def print_dict(init_dict, file_name="test"):
    """This function prints the initialization dict as a yaml file."""

    # Transfer the init dict in an ordered one to ensure that the init file is aligned
    #  appropriately
    ordered_dict = collections.OrderedDict()
    order = [
        "SIMULATION",
        "ESTIMATION",
        "TREATED",
        "UNTREATED",
        "CHOICE",
        "DIST",
        "VARTYPES",
        "SCIPY-BFGS",
        "SCIPY-POWELL",
    ]

    for key_ in order:
        if key_ in init_dict.keys():
            ordered_dict[key_] = init_dict[key_]
        else:
            pass
    for section in ["TREATED", "CHOICE", "UNTREATED", "DIST"]:
        if isinstance(ordered_dict[section]["params"], list):
            pass
        else:
            ordered_dict[section]["params"] = ordered_dict[section]["params"].tolist()

    # Print the initialization file
    with open(f"{file_name}.grmpy.yml", "w") as outfile:
        yaml.dump(
            ordered_dict,
            outfile,
            explicit_start=True,
            indent=4,
            width=99,
            default_flow_style=False,
        )



"""This module contains a tutorial illustrating the basic capabilities of the grmpy
package.
"""
import os

from grmpy.estimate.estimate import fit
from grmpy.simulate.simulate import simulate

f = os.path.dirname(__file__) + "/tutorial.grmpy.yml"
simulate(f)
rslt = fit(f)



"""The module contains the main function of the init file import process."""
import numpy as np
import yaml

from grmpy.check.check import check_presence_init
from grmpy.read.read_auxiliary import create_attr_dict_est, create_attr_dict_sim


def read(file, semipar=False, include_constant=False):
    """This function processes the initialization file
    for the estimation process.
    """
    # Check if there is an init file with the specified filename
    check_presence_init(file)

    # Load the initialization file
    with open(file) as y:
        init_dict = yaml.load(y, Loader=yaml.FullLoader)

    # If missing, add generic covariance matrix of the unobservables
    if semipar is False:
        try:
            init_dict["DIST"]
        except KeyError:
            init_dict["DIST"] = {"params": np.array([0.1, 0.0, 0.0, 0.1, 0.0, 1.0])}
        else:
            pass
    else:
        pass

    # Process the initialization file
    attr_dict = create_attr_dict_est(init_dict, semipar, include_constant)

    return attr_dict


def read_simulation(file):
    """Process initialization file for the simulation."""
    # Check if there is an init file with the specified filename
    check_presence_init(file)

    # Load the initialization file
    with open(file) as y:
        init_dict = yaml.load(y, Loader=yaml.FullLoader)

    # Process the initialization file
    attr_dict = create_attr_dict_sim(init_dict)

    return attr_dict


def check_append_constant(init_file, dict_, data, semipar=False):
    """Check if constant already provided by user.
    If not, add auto-generated constant.
    Pass in case there is a constant in first position of the data frame,
    but with a name other than 'const'.
    """
    if (
        "const" not in data
        and np.array_equal(np.asarray(data.iloc[:, 0]), np.ones(len(data))) is False
    ):
        dict_ = read(init_file, semipar, include_constant=True)
        data.insert(0, "const", 1.0)

    else:
        pass

    return dict_, data



"""This module contains test that check the code quality of the package."""
import os
import subprocess
from subprocess import CalledProcessError

# import pytest
from grmpy.grmpy_config import PACKAGE_DIR


# @pytest.mark.skip(reason="Obsolete. Would be used during development.")
def test_flake8():
    """Run flake8 to ensure the code quality. However, this is only relevant
    during development."""
    cwd = os.getcwd()
    os.chdir(PACKAGE_DIR)
    try:
        subprocess.check_call(["flake8"])
        os.chdir(cwd)
    except CalledProcessError:
        os.chdir(cwd)
        raise CalledProcessError(returncode=1, cmd="Flake8 standards not met.")



#!/usr/bin/env python

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import os
import sys

from pathlib import Path
from setuptools import Command, find_packages, setup
from shutil import rmtree

# Package meta-data.
NAME = "grmpy"
DESCRIPTION = (
    "grmpy is a Python package for the simulation and estimation of the "
    "generalized Roy model."
)
README = Path("README.md").read_text()
URL = "http://grmpy.readthedocs.io"
EMAIL = "eisenhauer@policy-lab.org"
AUTHOR = "The grmpy Development Team"

# What packages are required for this module to be executed?
REQUIRED = [
    "kernreg",
    "linearmodels",
    "numpy",
    "matplotlib",
    "oyaml",
    "pandas",
    "pytest",
    "scikit-learn",
    "scikit-misc",
    "scipy",
    "seaborn",
    "statsmodels",
]

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
# with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
#     long_description = '\n' + f.read()

# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, NAME, "__version__.py")) as f:
    exec(f.read(), about)


class PublishCommand(Command):
    """Support setup.py publish."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print(f"\033[1m{s}\033[0m")

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except FileNotFoundError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system(f"{sys.executable} setup.py sdist bdist_wheel --universal")

        self.status("Uploading the package to PyPi via Twine…")
        os.system("twine upload dist/*")

        sys.exit()


setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=DESCRIPTION + "\n\n" + README,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=find_packages(exclude=("tests",)),
    install_requires=REQUIRED,
    license="MIT",
    include_package_data=True,
    cmdclass={"publish": PublishCommand},
)
